#!/usr/bin/perl
# Process assembly listing into more readable format

use strict;
use warnings;

my %calls = (
  "0x8048134" => "main",
  "0x8048ecc" => "sendpackets",
  "0x8048f94" => "sendpacket",
  "0x8049138" => "get_addr",
  "0x8049174" => "dos1",
  "0x8049564" => "dos2",
  "0x80499f4" => "dos3",
  "0x8049d40" => "dos4",
  "0x804a194" => "encode",
  "0x804a1e8" => "decode",
  "0x804a2a8" => "setenv",
  "0x804a48c" => "unsetenv",
  "0x804bf80" => "gethostbyname",
  "0x804ce8c" => "inet_addr",
  "0x804f540" => "fclose",
  "0x804f620" => "fopen",
  "0x804f6d4" => "fread",
  "0x804f808" => "sprintf",
  "0x80555b0" => "usleep",
  "0x80555fc" => "execl",
  "0x80556cc" => "sleep",
  "0x80557e8" => "system",
  "0x80559a0" => "srand",
  "0x8055e38" => "rand",
  "0x8055fbc" => "exit",
  "0x8056058" => "rand",
  "0x8056480" => "bcopy",
  "0x805652c" => "memcpy",
  "0x80569bc" => "signal",
  "0x8056a2c" => "accept",
  "0x8056a74" => "bind",
  "0x8056b04" => "listen",
  "0x8056b44" => "recv",
  "0x8056b90" => "recvfrom",
  "0x8056bf0" => "send",
  "0x8056c3c" => "sendto",
  "0x8056c9c" => "setsockopt",
  "0x8056cf4" => "socket",
  "0x8057134" => "chdir",
  "0x8057160" => "close",
  "0x805718c" => "dup2",
  "0x80571e8" => "fork",
  "0x805720c" => "geteuid16",
  "0x80572b0" => "kill",
  "0x805733c" => "setsid",
  "0x80573bc" => "unlink",
  "0x8057444" => "time",
  "0x8057554" => "_exit",
  "0x8057764" => "memset",
  "0x805bd74" => "malloc",
  "0x805c290" => "free",
);

my %pass_ops = (
  "cld"        => 1,
  "cltd"       => 1,
  "int"        => 1,
  "movsb"      => 1,
  "movsw"      => 1,
  "repnz_scas" => 1,
  "repz_cmpsb" => 1,
  "repz_movsl" => 1,
  "ret"        => 1,
  "ror"        => 1,
  "xchg"       => 1,
);

my %comp_tbl = (
  "e"  => "==",
  "ne" => "!=",
  "a"  => "u>",
  "ae" => "u>=",
  "b"  => "u<",
  "be" => "u<=",
  "g"  => "s>",
  "ge" => "s>=",
  "l"  => "s<",
  "le" => "s<=",
  "s"  => "<0",
  "ns" => ">=0",
);

my %size_tag = (
  ""  => "",
  "b" => "byte",
  "w" => "word",
  "l" => "long",
);

my (%lastline, %currline);
sub lastval { exists($lastline{$_[0]}) ? $lastline{$_[0]} : $_[0] }

while (<>) {
  tweakargs();
  rewrite();
  print;
  %lastline = %currline;
  undef %currline;
}

sub rewrite {
  my @line = split /[\t\n]/;
  return unless @line == 3;
  my @op = split " ", $line[2] or return;
  my (@args, @oldargs);
  if (@op > 1) {
    @oldargs = @args = split /,/, pop @op;
    foreach (@args) {
      if (/^\$(0x[0-9a-fA-F]+)$/)  { $_ = $1 }
      elsif (/^(0x[0-9a-fA-F]+)$/) { $_ = "($1)" }
    }
  }
  my $op = join "_", @op;
  my $new;

  if ($op eq "shl" && @args == 2) { $new=assign($args[1], "<<", $args[0])}
  elsif ($op eq "shr" && @args == 2) {$new=assign($args[1], ">>>", $args[0])}
  elsif ($op eq "sar" && @args == 2) {$new=assign($args[1], ">>", $args[0])}
  elsif ($op eq "and" && @args == 2) {$new=assign($args[1], "&", $args[0])}
  elsif ($op eq "or" && @args == 2) {$new=assign($args[1], "|", $args[0])}
  elsif ($op eq "nop" && @args == 0) { $new = "" }
  elsif ($op eq "jmp" && @args == 1) { $new = "goto $oldargs[0]" }
  elsif ($op =~ /^mov([bwl]?)$/ && @args == 2) {
    my $targ = $size_tag{$1} . $args[1];
    my $val  = lastval($args[0]);
    $new = "$targ = $val";
    $currline{$targ} = "{$val}";
  }
  elsif ($op =~ /^mov([sz])([bw])l$/ && @args == 2) {
    my ($sign, $size) = ($1, $2);
    if ($sign eq "z") { $sign = "u" }
    my $targ = $args[1];
    my $val  = $sign . $size_tag{$size} . lastval($args[0]);
    $new = "$targ = $val";
    $currline{$targ} = "{$val}";
  }
  elsif ($op =~ /^add([bwl]?)$/ && @args == 2) {
    my $size = $1;
    my $val  = $args[0];
    if ($val =~ /^0x[0-9a-fA-F]+/ && (hex($val) & 0x80000000)) {
      $val = sprintf "0x%x", 1+~hex($val);
      $new = assign($args[1], "-", $val, $size);
    }
    else { $new = assign($args[1], "+", $val, $size) }
  }
  elsif ($op =~ /^sub([bwl]?)$/ && @args == 2) {
    $new = assign($args[1], "-", $args[0], $1);
  }
  elsif ($op =~ /^inc([bwl]?)$/ && @args == 1) {
    my $targ = $size_tag{$1} . $args[0];
    $new = "++$targ";
    my $val = lastval($targ);
    $currline{$targ} = "{$val+1}";
  }
  elsif ($op =~ /^dec([bwl]?)$/ && @args == 1) {
    my $targ = $size_tag{$1} . $args[0];
    $new = "--$targ";
    my $val = lastval($targ);
    $currline{$targ} = "{$val-1}";
  }
  elsif ($op =~ /^xor([bwl]?)$/ && @args == 2) {
    if ($args[0] eq $args[1]) {
      my $targ = $size_tag{$1} . $args[0];
      $currline{$targ} = "{0x0}";
      $new = "$targ = 0x0";
    }
    else { $new = assign($args[1], "^", $args[0], $1) }
  }
  elsif ($op =~ /^cmp([bwl]?)$/ && @args == 2) {
    my $lhs = lastval($size_tag{$1} . $args[1]);
    my $rhs = lastval($args[0]);
    $new = "cmp $lhs,$rhs";
    $currline{lhs} = $lhs;
    $currline{rhs} = $rhs;
  }
  elsif ($op eq "test" && @args == 2) {
    my $lhs = lastval($args[1]);
    my $rhs = lastval($args[0]);
    $new = "test $lhs,$rhs";
    if ($lhs eq $rhs) {
      $currline{lhs} = $lhs;
      $currline{rhs} = "0x0";
    }
  }
  elsif ($op eq "idiv" && @args == 1) {
    $new = "idiv $args[0]";
    $currline{'%eax'} = "{quot}";
    $currline{'%edx'} = "{rem}";
  }
  elsif ($op eq "neg" && @args == 1) {
    my $targ = $args[0];
    my $val  = lastval($targ);
    $new = "$targ = -$val";
    $currline{$targ} = "{-$val}";
  }
  elsif ($op eq "not" && @args == 1) {
    my $targ = $args[0];
    my $val  = lastval($targ);
    $new = "$targ = ~$val";
    $currline{$targ} = "{~$val}";
  }
  elsif (($op eq "push" || $op eq "pop") && @args == 1) {
    my $val = lastval($args[0]);
    $new = "$op $val";
  }
  elsif ($op eq "lea" && @args == 2) {
    my $targ = $args[1];
    if ($args[0] =~ /^\(([^()]+)\)$/) {
      if ($1 eq $targ) { $new = "" }
      else {
        $new = "$targ = $1";
        $currline{$targ} = "{$1}";
      }
    }
  }
  elsif ($op eq "call" && @args == 1) {
    my $func = $calls{$oldargs[0]};
    $new = "call $func" if defined $func;
  }
  elsif ($op =~ /^j(.*)/ && $comp_tbl{$1} && @args == 1) {
    my $test = gettest($1);
    $new = "if ($test) goto $oldargs[0]";
  }
  elsif ($op =~ /^set(.*)/ && $comp_tbl{$1} && @args == 1) {
    my $test = gettest($1);
    $new = "$args[0] = $test ? 1 : 0";
  }
  elsif ($pass_ops{$op}) {
    $new = $op . " " . join(",", reverse @args);
  }

  $_ = "$line[0]\t; $new\n" if defined $new;
} # rewrite

sub assign {
  my ($targ, $op, $val, $size) = @_;
  $size ||= "";
  $size = $size_tag{$size} || $size;
  $targ = $size . $targ;
  $val  = $lastline{$val} if exists $lastline{$val};
  my $tval = lastval($targ);
  $currline{$targ} = "{$tval$op$val}";
  return "$targ $op= $val";
} # assign

sub gettest {
  my ($cmp) = @_;
  my $lhs = $lastline{lhs} || "";
  my $rhs = $lastline{rhs} || "";
  %currline = %lastline;
  return "$lhs $comp_tbl{$cmp} $rhs";
} # gettest

sub tweakargs {
  s{([\s,])
    (0x[0-9a-fA-F]+)?
    \(([^()]+)\)
    ([\s,])}
  {
    my ($pre, $const, $reg, $post, $new) = ($1, $2, $3, $4);
    my @regs = split /,/, $reg;
    if (@regs == 2) { unshift @regs, "" }
    if (@regs == 3) {
      $new  = "";
      $new  = "$regs[0]+" unless $regs[0] eq "";
      $new .= $regs[1];
      $new .= "*$regs[2]" unless $regs[2] eq "1";
    }
    else {
      $new = join "+", @regs;
    }
    if (defined $const) {
      my $val = hex($const);
      if ($val & 0x80000000) { $new .= sprintf "-0x%x", 1+~hex($const) }
      elsif ($val != 0)      { $new .= "+$const" }
    }
    "$pre($new)$post"
  }egx;
} # parseop

